from .scalar import (
    Typeless,
    Null,
    Boolean,
    Number,
    String
)
from .array import List, Tuple
from .object import Object

GENERATORS = (
    Null,
    Boolean,
    Number,
    String,
    List,
    Tuple,
    Object
)

__all__ = tuple(list(GENERATORS) + [Typeless, GENERATORS])



from .generators import GENERATORS, Typeless


class SchemaGenerationError(RuntimeError):
    pass


class SchemaNode(object):
    """
    Basic schema generator class. SchemaNode objects can be loaded
    up with existing schemas and objects before being serialized.
    """
    generator_classes = GENERATORS

    def __init__(self):
        self._schema_generators = []

    def add_schema(self, schema):
        """
        Merges in an existing schema.

        arguments:
        * `schema` (required - `dict` or `SchemaNode`):
          an existing JSON Schema to merge.
        """

        # serialize instances of SchemaNode before parsing
        if isinstance(schema, SchemaNode):
            schema = schema.to_schema()

        for subschema in self._get_subschemas(schema):
            # delegate to SchemaType object
            schema_generator = self._get_generator_for_schema(subschema)
            schema_generator.add_schema(subschema)

        # return self for easy method chaining
        return self

    def add_object(self, obj):
        """
        Modify the schema to accomodate an object.

        arguments:
        * `obj` (required - `dict`):
          a JSON object to use in generating the schema.
        """

        # delegate to SchemaType object
        schema_generator = self._get_generator_for_object(obj)
        schema_generator.add_object(obj)

        # return self for easy method chaining
        return self

    def to_schema(self):
        """
        Convert the current schema to a `dict`.
        """
        types = set()
        generated_schemas = []
        for schema_generator in self._schema_generators:
            generated_schema = schema_generator.to_schema()
            if len(generated_schema) == 1 and 'type' in generated_schema:
                types.add(generated_schema['type'])
            else:
                generated_schemas.append(generated_schema)

        if types:
            if len(types) == 1:
                (types,) = types
            else:
                types = sorted(types)
            generated_schemas = [{'type': types}] + generated_schemas
        if len(generated_schemas) == 1:
            (result_schema,) = generated_schemas
        elif generated_schemas:
            result_schema = {'anyOf': generated_schemas}
        else:
            result_schema = {}

        return result_schema

    def __len__(self):
        return len(self._schema_generators)

    def __eq__(self, other):
        # TODO: find a more optimal way to do this
        if self is other:
            return True
        if not isinstance(other, type(self)):
            return False

        return self.to_schema() == other.to_schema()

    def __ne__(self, other):
        return not self.__eq__(other)

    # private methods

    def _get_subschemas(self, schema):
        if 'anyOf' in schema:
            return schema['anyOf']
        elif isinstance(schema.get('type'), list):
            other_keys = dict(schema)
            del other_keys['type']
            return [dict(type=tipe, **other_keys) for tipe in schema['type']]
        else:
            return [schema]

    def _get_generator_for_schema(self, schema):
        return self._get_generator_for_('schema', schema)

    def _get_generator_for_object(self, obj):
        return self._get_generator_for_('object', obj)

    def _get_generator_for_(self, kind, schema_or_obj):
        # check existing types
        for schema_generator in self._schema_generators:
            if getattr(schema_generator, 'match_' + kind)(schema_or_obj):
                return schema_generator

        # check all potential types
        for schema_generator_class in self.generator_classes:
            if getattr(schema_generator_class, 'match_' + kind)(schema_or_obj):
                schema_generator = schema_generator_class(type(self))

                # incorporate typeless generator if it exists
                if self._schema_generators and \
                        isinstance(self._schema_generators[-1], Typeless):
                    typeless = self._schema_generators.pop()
                    schema_generator.add_schema(typeless.to_schema())

                self._schema_generators.append(schema_generator)
                return schema_generator

        # no match found, if typeless add to first generator
        if kind == 'schema' and Typeless.match_schema(schema_or_obj):
            if not self._schema_generators:
                self._schema_generators.append(Typeless(type(self)))
            schema_generator = self._schema_generators[0]
            return schema_generator

        # no match found, raise an error
        raise SchemaGenerationError(
            'Could not find matching type for {0}: {1!r}'.format(
                kind, schema_or_obj))



import json
from warnings import warn
from .node import SchemaNode


class SchemaBuilder(object):
    """
    ``SchemaBuilder`` is the basic schema generator class.
    ``SchemaBuilder`` instances can be loaded up with existing schemas
    and objects before being serialized.
    """
    DEFAULT_URI = 'http://json-schema.org/schema#'
    NULL_URI = 'NULL'

    def __init__(self, schema_uri='DEFAULT'):
        """
        :param schema_uri: value of the ``$schema`` keyword. If not
          given, it will use the value of the first available
          ``$schema`` keyword on an added schema or else the default:
          ``'http://json-schema.org/schema#'``. A value of ``False`` or
          ``None`` will direct GenSON to leave out the ``"$schema"``
          keyword.
        """
        if schema_uri is None or schema_uri is False:
            self.schema_uri = self.NULL_URI
        elif schema_uri == 'DEFAULT':
            self.schema_uri = None
        else:
            self.schema_uri = schema_uri

        self._root_node = SchemaNode()

    def add_schema(self, schema):
        """
        Merge in a JSON schema. This can be a ``dict`` or another
        ``SchemaBuilder``

        :param schema: a JSON Schema

        .. note::
            There is no schema validation. If you pass in a bad schema,
            you might get back a bad schema.
        """
        if isinstance(schema, SchemaBuilder):
            schema_uri = schema.schema_uri
            schema = schema.to_schema()
            if schema_uri is None:
                del schema['$schema']
        elif isinstance(schema, SchemaNode):
            schema = schema.to_schema()

        if '$schema' in schema:
            self.schema_uri = self.schema_uri or schema['$schema']
            schema = dict(schema)
            del schema['$schema']
        self._root_node.add_schema(schema)

    def add_object(self, obj):
        """
        Modify the schema to accomodate an object.

        :param obj: any object or scalar that can be serialized in JSON
        """
        self._root_node.add_object(obj)

    def to_schema(self):
        """
        Generate a schema based on previous inputs.

        :rtype: ``dict``
        """
        schema = self._base_schema()
        schema.update(self._root_node.to_schema())
        return schema

    def to_json(self, *args, **kwargs):
        """
        Generate a schema and convert it directly to serialized JSON.

        :rtype: ``str``
        """
        return json.dumps(self.to_schema(), *args, **kwargs)

    def __len__(self):
        """
        Number of ``SchemaGenerator``s at the top level. This is used
        mostly to check for emptiness.
        """
        return len(self._root_node)

    def __eq__(self, other):
        """
        Check for equality with another ``SchemaBuilder`` object.

        :param other: another ``SchemaBuilder`` object. Other types are
          accepted, but will always return ``False``
        """
        if self is other:
            return True
        if not isinstance(other, SchemaBuilder):
            return False

        return self._root_node == other._root_node

    def __ne__(self, other):
        return not self.__eq__(other)

    def _base_schema(self):
        if self.schema_uri == self.NULL_URI:
            return {}
        else:
            return {'$schema': self.schema_uri or self.DEFAULT_URI}


class Schema(SchemaBuilder):

    def __init__(self):
        warn('genson.Schema is deprecated in v1.0, and it may be '
             'removed in future versions. Use genson.SchemaBuilder'
             'instead.',
             PendingDeprecationWarning)
        super(Schema, self).__init__(schema_uri=SchemaBuilder.NULL_URI)

    def to_dict(self, recurse='DEPRECATED'):
        warn('#to_dict is deprecated in v1.0, and it may be removed in '
             'future versions. Use #to_schema instead.',
             PendingDeprecationWarning)
        if recurse != 'DEPRECATED':
            warn('the `recurse` option for #to_dict does nothing in v1.0',
                 DeprecationWarning)
        return self.to_schema()
